#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Client module for connecting to an rpc_server in a remote machine

Test of execution:
python rpc_client.py http://localhost:8000/patata popo_method int:3.5  float:5.6
"""

import sys
import xmlrpclib

DEBUG = True

_cmd_struct = {
    'endpoint': None,
    'method_name': None,
    'args': []
}


def parse_cmd_line(args):

    admitted_types = ('int', 'float', 'str')

    cmd_parts = dict(_cmd_struct)
    cmd_parts['endpoint'] = args[1]
    cmd_parts['method_name'] = args[2]
    cmd_parts['args'] = []

    raw_method_args = args[3:]

    for raw_arg in raw_method_args:

        #arg_parts = raw_arg.strip('"\'').split(':')
        arg_parts = raw_arg.split(':')

        if len(arg_parts) == 1 or arg_parts[0] not in admitted_types:
            arg_type, arg_value = 'str', ':'.join(arg_parts[0:])
            if arg_value.isdigit():
                arg_type = 'int'
            elif not (arg_value.startswith('.') or arg_value.endswith('.')) and arg_value.replace('.', '', 1).isdigit():
                arg_type = 'float'
        else:
            arg_type, arg_value = arg_parts[0], ':'.join(arg_parts[1:])

        try:
            value = eval('{0}({1})'.format(arg_type, arg_value)) if arg_type != 'str' else arg_value
        except Exception:
            print >> sys.stderr, "\n Error: Detected argument with wrong format"
            sys.exit(3)

        cmd_parts['args'].append(value)

    return cmd_parts


def get_rpc_client(endpoint):
    """
    Get an rpc client object to remote server listening at endpoint
    :param endpoint: http://host:port/rpc_path
    :return: rpc_client object
    """
    return xmlrpclib.ServerProxy(endpoint)


if __name__ == '__main__':

    if len(sys.argv) <= 2:
        print >>sys.stderr, 'usage: {0} http://<host>:<port>/<rpc_path> <method_name> [ [int|float|str]:arg1 ' \
                            '[int|float|str]:arg2 ...[int|float|str]:argN ...]'.format(sys.argv[0])
        sys.exit(1)

    cmd_line = parse_cmd_line(sys.argv[:])

    if DEBUG:
        print 'Parsed cmd_line: ', cmd_line

    client = get_rpc_client(cmd_line['endpoint'])

    #Executing now the remote method
    result = getattr(client, cmd_line['method_name'])(*cmd_line['args'])
    if result is not None:
        print ">>Result:\n", result



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import socket
from contextlib import closing

from zmon_worker_monitor.adapters.ifunctionfactory_plugin import IFunctionFactoryPlugin, propartial


class TcpFactory(IFunctionFactoryPlugin):

    def __init__(self):
        super(TcpFactory, self).__init__()

    def configure(self, conf):
        """
        Called after plugin is loaded to pass the [configuration] section in their plugin info file
        :param conf: configuration dictionary
        """
        return

    def create(self, factory_ctx):
        """
        Automatically called to create the check function's object
        :param factory_ctx: (dict) names available for Function instantiation
        :return: an object that implements a check function
        """
        return propartial(TcpWrapper, host=factory_ctx['host'])


class TcpWrapper(object):

    def __init__(self, host, timeout=10):
        self.host = host
        self.timeout = timeout

    def open(self, *ports):
        results = {}
        for port in ports:
            with closing(socket.socket()) as s:
                s.settimeout(self.timeout)
                try:
                    s.connect((self.host, port))
                except Exception, e:
                    results[port] = str(e)
                else:
                    results[port] = 'OK'
        if len(ports) == 1:
            return results[ports[0]]
        else:
            return results

    def resolve(self, host):
        try:
            result = socket.gethostbyname(host)
        except Exception, e:
            result = 'ERROR: ' + str(e)
        return result


if __name__ == '__main__':
    import sys
    tcp = TcpWrapper(sys.argv[1])
    print tcp.open(22, 80, 123)
    print tcp.open(22)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from datetime import timedelta, datetime
import re

TIME_UNITS = {
    's': 'seconds',
    'm': 'minutes',
    'h': 'hours',
    'd': 'days',
}

TIME_FORMATS = ['%Y-%m-%d %H:%M:%S.%f', '%Y-%m-%d %H:%M:%S', '%Y-%m-%d %H:%M', '%Y-%m-%d']

TIMEZONE_OFFSET = re.compile(r'([+-])([0-9][0-9])(?::?([0-9][0-9]))?$')


def parse_timedelta(s):
    '''
    >>> parse_timedelta('bla')


    >>> parse_timedelta('1s').total_seconds()
    1.0

    >>> parse_timedelta('-2s').total_seconds()
    -2.0

    >>> parse_timedelta('2m').total_seconds()
    120.0

    >>> parse_timedelta('1h').total_seconds()
    3600.0
    '''

    if s.startswith('-'):
        s = s[1:]
        factor = -1
    else:
        factor = 1
    try:
        v = int(s[:-1])
        u = s[-1]
    except:
        return None

    arg = TIME_UNITS.get(u)
    if arg:
        return factor * timedelta(**{arg: v})
    return None


def parse_datetime(s):
    '''
    >>> parse_datetime('foobar')

    >>> parse_datetime('1983-10-12T23:30').isoformat(' ')
    '1983-10-12 23:30:00'

    >>> parse_datetime('1983-10-12 23:30:12').isoformat(' ')
    '1983-10-12 23:30:12'

    >>> parse_datetime('2014-05-05 17:40:44.100313').isoformat(' ')
    '2014-05-05 17:40:44.100313'

    >>> parse_datetime('2014-05-05 17:40:44.100313+01:00').isoformat(' ')
    '2014-05-05 16:40:44.100313'
    '''

    s = s.replace('T', ' ')

    # calculate timezone data from date string, we'll parse it ourselves ('%z' is not supported on all platforms for strptime)
    match = TIMEZONE_OFFSET.search(s)
    if match:
        signum = int(match.group(1) + '1')
        hours = signum * int(match.group(2))
        minutes = signum * int(match.group(3))
        timezone_timedelta = timedelta(hours=hours, minutes=minutes)
    else:
        timezone_timedelta = timedelta()

    # remove timezone data from input string, if any.
    s = TIMEZONE_OFFSET.sub('', s)
    for fmt in TIME_FORMATS:
        try:
            return datetime.strptime(s, fmt) - timezone_timedelta
        except:
            pass
    return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys
from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand

MAIN_PACKAGE = 'zmon_worker_monitor'


class PyTest(TestCommand):

    user_options = [('cov-html=', None, 'Generate junit html report')]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.cov = None
        self.pytest_args = ['--cov', MAIN_PACKAGE, '--cov-report', 'term-missing',
                            '--doctest-modules', '-s',
                            '--ignore', 'tests/plugins']
        self.cov_html = False

    def finalize_options(self):
        TestCommand.finalize_options(self)
        if self.cov_html:
            self.pytest_args.extend(['--cov-report', 'html'])

    def run_tests(self):
        try:
            import pytest
        except:
            raise RuntimeError('py.test is not installed, run: pip install pytest')

        # HACK: circumvent strange atexit error with concurrent.futures
        # https://developer.blender.org/T39399
        import concurrent.futures  # noqa

        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


def load_req(fn):
    return [r.strip() for r in open(fn).read().splitlines() if r.strip() and not r.strip().startswith('#')]

if __name__ == '__main__':
    # just in case setup.py is launched from elsewhere than the containing directory
    original_dir = os.getcwd()
    os.chdir(os.path.dirname(os.path.abspath(__file__)))

    try:
        cmdclass = {}
        cmdclass['test'] = PyTest

        setup(
            name="zmon-worker",
            version=__import__('zmon_worker_monitor').__version__,
            description='ZMON Worker Monitor',
            url='https://github.com/zalando/zmon-worker',
            license='Apache License 2.0',
            packages=find_packages(exclude=['tests', 'tests.*']),
            setup_requires=['numpy', 'flake8'],  # workaround for bug in numpy+setuptools: https://github.com/numpy/numpy/issues/2434
            install_requires=load_req('requirements.txt'),
            cmdclass=cmdclass,
            test_suite='tests',
            tests_require=load_req('test_requirements.txt'),

            entry_points={
                'console_scripts': [
                    'zmon-worker = zmon_worker_monitor.web:main',
                ]
            },

            package_data={
                'zmon_worker_monitor': ['data/*', 'builtins/plugins/*.worker_plugin'],

            },

            # more metadata for upload to PyPI
            author="Henning Jacobs",
            author_email="henning.jacobs@zalando.de",
            keywords='zalando zmon zmon2 worker component monitoring infrastructure',
            long_description=open('README.rst').read(),
            classifiers=[
                'Development Status :: 5 - Production/Stable',
                'Intended Audience :: Developers',
                'License :: OSI Approved :: Apache Software License',
                'Operating System :: OS Independent',
                'Programming Language :: Python',
                'Programming Language :: Python :: 2',
                'Topic :: Software Development :: Libraries :: Python Modules'],

            platforms='All',
        )

    finally:
        os.chdir(original_dir)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import subprocess

from zmon_worker_monitor.adapters.ifunctionfactory_plugin import IFunctionFactoryPlugin, propartial


class PingFactory(IFunctionFactoryPlugin):

    def __init__(self):
        super(PingFactory, self).__init__()

    def configure(self, conf):
        """
        Called after plugin is loaded to pass the [configuration] section in their plugin info file
        :param conf: configuration dictionary
        """
        return

    def create(self, factory_ctx):
        """
        Automatically called to create the check function's object
        :param factory_ctx: (dict) names available for Function instantiation
        :return: an object that implements a check function
        """
        return propartial(ping, host=factory_ctx['host'])


def ping(host, count=1, timeout=1):
    cmd = [
        'ping',
        '-c',
        str(count),
        '-w',
        str(timeout),
        host,
    ]

    sub = subprocess.Popen(cmd, shell=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
    sub.communicate()
    ret = sub.wait() == 0
    return ret


if __name__ == '__main__':
    import sys
    print ping(sys.argv[1])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Query Tomcat metrics via JMX
"""

import json
import sys

from zmon_worker_monitor.adapters.ifunctionfactory_plugin import IFunctionFactoryPlugin, propartial
from zmon_worker_monitor import plugin_manager


MILLI = 10 ** -3
NANO = 10 ** -9

THREAD_POOL_PORT_PREFIXES = {'http': 3, 'ajp': 2}


class ZomcatFactory(IFunctionFactoryPlugin):

    def __init__(self):
        super(ZomcatFactory, self).__init__()
        # fields to store dependencies: plugin depends on 3 other plugins
        self.http_factory = None
        self.jmx_factory = None
        self.counter_factory = None

    def configure(self, conf):
        """
        Called after plugin is loaded to pass the [configuration] section in their plugin info file
        :param conf: configuration dictionary
        """
        return

    def create(self, factory_ctx):
        """
        Automatically called to create the check function's object
        :param factory_ctx: (dict) names available for Function instantiation
        :return: an object that implements a check function
        """
        # load plugins dependencies and store them locally for efficiency
        if not self.http_factory:
            self.http_factory = plugin_manager.get_plugin_obj_by_name('http', 'Function')

        if not self.jmx_factory:
            self.jmx_factory = plugin_manager.get_plugin_obj_by_name('jmx', 'Function')

        if not self.counter_factory:
            self.counter_factory = plugin_manager.get_plugin_obj_by_name('counter', 'Function')

        return propartial(ZomcatWrapper,
                          host=factory_ctx['host'],
                          instance=factory_ctx['instance'],
                          http=self.http_factory.create(factory_ctx),
                          jmx=self.jmx_factory.create(factory_ctx),
                          counter=self.counter_factory.create(factory_ctx))


def memory_usage_percentage(data):
    '''helper function to calculate percentage from "used" and "max"'''

    return round(100. * data['used'] / data['max'], 2)


class ZomcatWrapper(object):

    def __init__(self, host, instance, http, jmx, counter):
        '''expects ready to use "partials" for http, jmx and counter'''

        self.host = host
        self.instance = instance
        self._http = http
        self._jmx = jmx
        # initialize counter with Redis connection
        self._counter = counter('')

    @staticmethod
    def get_jmx_port(instance):
        return int('4' + instance)

    def gc(self):
        '''return garbage collector statistics: gc_percentage and gcs_per_sec'''

        gc_count = 0
        gc_time = 0
        for _type in 'ConcurrentMarkSweep', 'ParNew':
            try:
                row = self._jmx().query('java.lang:type=GarbageCollector,name={}'.format(_type), 'CollectionCount',
                                        'CollectionTime').results()
                gc_count += row['CollectionCount']
                gc_time += row['CollectionTime']
            except:
                pass

        gc_count = round(self._counter.key('gcCount').per_second(gc_count), 2)
        gc_time = self._counter.key('gcTime').per_second(gc_time * MILLI)
        return {'gc_percentage': round(gc_time * 100, 2), 'gcs_per_sec': gc_count}

    def requests(self):
        '''return Tomcat request statistics such as requests/s and errors/s'''

        request_count = 0
        request_time = 0
        http_errors = 0
        for _type, _port_prefix in THREAD_POOL_PORT_PREFIXES.items():
            try:
                row = self._jmx().query('Catalina:type=GlobalRequestProcessor,name="{}-apr-{}{}"'.format(_type,
                                        _port_prefix, self.instance), 'requestCount', 'processingTime', 'errorCount'
                                        ).results()
                request_count += row['requestCount']
                request_time += row['processingTime']
                http_errors += row['errorCount']
            except:
                pass
        requests = round(self._counter.key('requestCount').per_second(request_count), 2)
        http_errors = round(self._counter.key('errorCount').per_second(http_errors), 2)
        time_per_request = round(self._counter.key('requestTime').per_second(request_time) / max(requests, 1), 2)

        return {'http_errors_per_sec': http_errors, 'requests_per_sec': requests, 'time_per_request': time_per_request}

    def basic_stats(self):
        '''return basic statistics such as memory, CPU and thread usage'''

        jmx = self._jmx()
        # {"NonHeapMemoryUsage":{"max":184549376,"init":24313856,"used":54467720,"committed":85266432},
        # "HeapMemoryUsage":{"max":518979584,"init":134217728,"used":59485272,"committed":129761280}}
        jmx.query('java.lang:type=Memory', 'HeapMemoryUsage', 'NonHeapMemoryUsage')
        jmx.query('java.lang:type=Threading', 'ThreadCount')
        jmx.query('java.lang:type=OperatingSystem', 'ProcessCpuTime')
        data = jmx.results()
        memory = data['java.lang:type=Memory']
        threading = data['java.lang:type=Threading']
        os = data['java.lang:type=OperatingSystem']
        cpu = self._counter.key('cpuTime').per_second(os['ProcessCpuTime'] * NANO)
        threads = threading['ThreadCount']

        try:
            heartbeat = self._http('/heartbeat.jsp', timeout=3).text().strip() == 'OK: Zalando JVM is running'
        except:
            heartbeat = None
        try:
            jobs = self._http('/jobs.monitor?view=json', timeout=3).json()['operationMode'] == 'NORMAL'
        except:
            jobs = None

        return {
            'cpu_percentage': round(cpu * 100, 2),
            'heap_memory_percentage': memory_usage_percentage(memory['HeapMemoryUsage']),
            'heartbeat_enabled': heartbeat,
            'jobs_enabled': jobs,
            'nonheap_memory_percentage': memory_usage_percentage(memory['NonHeapMemoryUsage']),
            'threads': threads,
        }

    def health(self):
        '''return complete Zomcat health statistics including memory, threads, CPU, requests and GC'''

        data = {}
        data.update(self.basic_stats())
        data.update(self.gc())
        data.update(self.requests())
        return data


if __name__ == '__main__':

    # init plugin manager and collect plugins, as done by Zmon when worker is starting
    plugin_manager.init_plugin_manager()
    plugin_manager.collect_plugins(load_builtins=True, load_env=True)

    host = sys.argv[1]
    instance = sys.argv[2]

    factory_ctx = {
        'base_url': 'http://{host}:3{instance}/'.format(host=host, instance=instance),
        'host': host,
        'port': ZomcatWrapper.get_jmx_port(instance),
        'instance': instance,
        'redis_host': 'localhost',
    }

    # http = partial(HttpWrapper, base_url='http://{host}:3{instance}/'.format(host=host, instance=instance))
    http = plugin_manager.get_plugin_obj_by_name('http', 'Function').create(factory_ctx)

    # jmx = partial(JmxWrapper, host=host, port=ZomcatWrapper.get_jmx_port(instance))
    jmx = plugin_manager.get_plugin_obj_by_name('jmx', 'Function').create(factory_ctx)

    # counter = partial(CounterWrapper, redis_host='localhost')
    counter = plugin_manager.get_plugin_obj_by_name('counter', 'Function').create(factory_ctx)

    zomcat = ZomcatWrapper(host, instance, http=http, jmx=jmx, counter=counter)
    print json.dumps(zomcat.health(), indent=4, sort_keys=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import requests
import sys
import urllib
import urlparse
import json
from prometheus_client.parser import text_string_to_metric_families
from collections import defaultdict

from zmon_worker_monitor.zmon_worker.errors import HttpError
from requests.adapters import HTTPAdapter

from zmon_worker_monitor.adapters.ifunctionfactory_plugin import IFunctionFactoryPlugin, propartial

import tokens

# will use OAUTH2_ACCESS_TOKEN_URL environment variable by default
# will try to read application credentials from CREDENTIALS_DIR
tokens.configure()
tokens.manage('uid', ['uid'])
tokens.start()

logger = logging.getLogger('zmon-worker.http-function')

class HttpFactory(IFunctionFactoryPlugin):

    def __init__(self):
        super(HttpFactory, self).__init__()

    def configure(self, conf):
        """
        Called after plugin is loaded to pass the [configuration] section in their plugin info file
        :param conf: configuration dictionary
        """
        return

    def create(self, factory_ctx):
        """
        Automatically called to create the check function's object
        :param factory_ctx: (dict) names available for Function instantiation
        :return: an object that implements a check function
        """
        return propartial(HttpWrapper, base_url=factory_ctx.get('entity_url'))


def absolute_http_url(url):
    '''
    >>> absolute_http_url('')
    False

    >>> absolute_http_url('bla:8080/blub')
    False

    >>> absolute_http_url('https://www.zalando.de')
    True
    '''

    return url.startswith('http://') or url.startswith('https://')


class HttpWrapper(object):

    def __init__(
        self,
        url,
        params=None,
        base_url=None,
        timeout=10,
        max_retries=0,
        verify=True,
        oauth2=False,
        headers=None,
    ):

        self.url = (base_url + url if not absolute_http_url(url) else url)
        self.clean_url = None
        self.params = params
        self.timeout = timeout
        self.max_retries = max_retries
        self.verify = verify
        self.headers = headers or {}
        self.oauth2 = oauth2
        self.__r = None

    def __request(self, raise_error=True, post_data = None):
        if self.__r is not None:
            return self.__r
        if self.max_retries:
            s = requests.Session()
            s.mount('', HTTPAdapter(max_retries=self.max_retries))
        else:
            s = requests

        base_url = self.url
        basic_auth = None

        url_parsed = urlparse.urlsplit(base_url)
        if url_parsed and url_parsed.username and url_parsed.password:
            base_url = base_url.replace("{0}:{1}@".format(urllib.quote(url_parsed.username), urllib.quote(url_parsed.password)), "")
            base_url = base_url.replace("{0}:{1}@".format(url_parsed.username, url_parsed.password), "")
            basic_auth = requests.auth.HTTPBasicAuth(url_parsed.username, url_parsed.password)
        self.clean_url = base_url

        if self.oauth2:
            self.headers.update({'Authorization':'Bearer {}'.format(tokens.get('uid'))})

        try:
            if post_data is None:
                self.__r = s.get(base_url, params=self.params, timeout=self.timeout, verify=self.verify,
                                 headers=self.headers, auth = basic_auth)
            else:
                self.__r = s.post(base_url, params=self.params, timeout=self.timeout, verify=self.verify,
                                  headers=self.headers, auth = basic_auth, data=json.dumps(post_data))
        except requests.Timeout, e:
            raise HttpError('timeout', self.clean_url), None, sys.exc_info()[2]
        except requests.ConnectionError, e:
            raise HttpError('connection failed', self.clean_url), None, sys.exc_info()[2]
        except Exception, e:
            raise HttpError(str(e), self.clean_url), None, sys.exc_info()[2]
        if raise_error:
            try:
                self.__r.raise_for_status()
            except requests.HTTPError, e:
                raise HttpError(str(e), self.clean_url), None, sys.exc_info()[2]
        return self.__r

    def json(self, raise_error=True):
        r = self.__request(raise_error=raise_error)
        try:
            return r.json()
        except Exception, e:
            raise HttpError(str(e), self.url), None, sys.exc_info()[2]

    def jolokia(self, read_requests, raise_error=True):

        def set_read_type(x):
            x['type'] = 'READ'

        # hack quick verify
        if (not self.url.endswith('jolokia/')) or ('?' in self.url) or ('&' in self.url):
            raise HttpError("URL needs to end in jolokia/ and not contain ? and &", self.url)

        map(set_read_type, read_requests)

        r = self.__request(post_data=read_requests, raise_error=raise_error)

        try:
            return r.json()
        except Exception, e:
            raise HttpError(str(e), self.url), None, sys.exc_info()[2]


    def actuator_metrics(self, prefix = 'zmon.response.', raise_error = True):
        """
        /metric responds with keys like: zmon.response.<status>.<method>.<end-point>

        Response map is ep->method->status->metric
        """
        j = self.json(raise_error=raise_error)
        r={}

        # for clojure projects we use the dropwizard servlet, there the json looks slightly different
        if "timers" in j:
            metric_map = {'p99':'99th','p75':'75th','mean':'median','m1_rate':'mRate','99%':'99th','75%':'75th','1m.rate':'mRate'}
            j = j["timers"]
            j["zmon.response.200.GET.metrics"]={"mRate": 0.12345}

            start_index = len(prefix.split('.')) - 1

            for (k,v) in j.iteritems():
                if k.startswith(prefix):
                    ks = k.split('.')
                    ks = ks[start_index:]

                    status = ks[0]
                    method = ks[1]
                    ep = '.'.join(ks[2:])

                    if not ep in r:
                        r[ep]={}

                    if not method in r[ep]:
                        r[ep][method]={}

                    if not status in r[ep][method]:
                        r[ep][method][status]={}

                    for (mn, mv) in v.iteritems():
                        if mn in ['count','p99','p75','m1_rate','min','max','mean','75%','99%','1m.rate','median']:
                            if mn in metric_map:
                                mn = metric_map[mn]
                            r[ep][method][status][mn]=mv
            return r

        j["zmon.response.200.GET.metrics.oneMinuteRate"]=0.12345
        for (k,v) in j.iteritems():
            if k.startswith(prefix):
                ks = k.split('.')

                if ks[-2]=='snapshot':
                    ep = '.'.join(ks[4:-2])
                else:
                    ep = '.'.join(ks[4:-1])

                if not ep in r:
                    r[ep] = {}

                # zmon.response. 200 . GET . EP .

                if ks[3] not in r[ep]:
                    r[ep][ks[3]] = {}

                if ks[2] not in r[ep][ks[3]]:
                    r[ep][ks[3]][ks[2]] = {}

                if not (ks[-2] == 'snapshot'):
                    if ks[-1] == 'count':
                        r[ep][ks[3]][ks[2]]['count']=v
                    if ks[-1] == 'oneMinuteRate':
                        r[ep][ks[3]][ks[2]]['mRate']=v
                else:
                    if ks[-1] in ['75thPercentile','99thPercentile','min','max','median']:
                        r[ep][ks[3]][ks[2]][ks[-1].replace("Percentile", "")] = v

        return r

    def text(self, raise_error=True):
        r = self.__request(raise_error=raise_error)
        return r.text

    def prometheus(self):
        t = self.__request().text
        samples_by_name = defaultdict(list)

        for l in text_string_to_metric_families(t):
            for s in l.samples:
                samples_by_name[s[0]].append((s[1],s[2]))

        return samples_by_name

    def headers(self, raise_error=True):
        return self.__request(raise_error=raise_error).headers

    def cookies(self, raise_error=True):
        return self.__request(raise_error=raise_error).cookies

    def content_size(self, raise_error=True):
        return len(self.__request(raise_error=raise_error).content)

    def time(self, raise_error=True):
        return self.__request(raise_error=raise_error).elapsed.total_seconds()

    def code(self):
        return self.__request(raise_error=False).status_code


if __name__ == '__main__':
    http = HttpWrapper(sys.argv[1], max_retries=3)
    print http.text()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Zalando-specific function to monitor job locking (Job Framework uses Redis to lock)
"""

try:
    from cmdb.client import Client as cmdb_client
except:
    cmdb_client = None
from dogpile.cache import make_region

from zmon_worker_monitor.adapters.ifunctionfactory_plugin import IFunctionFactoryPlugin, propartial

import json
import redis
import time

HOSTS_CACHE_EXPIRATION_TIME = 600  # 10 minutes
memory_cache = make_region().configure('dogpile.cache.memory', expiration_time=HOSTS_CACHE_EXPIRATION_TIME)


class JoblocksFactory(IFunctionFactoryPlugin):

    def __init__(self):
        super(JoblocksFactory, self).__init__()
        # fields from configuration
        self.cmdb_url = None

    def configure(self, conf):
        """
        Called after plugin is loaded to pass the [configuration] section in their plugin info file
        :param conf: configuration dictionary
        """
        self.cmdb_url = conf['cmdb_url']

    def create(self, factory_ctx):
        """
        Automatically called to create the check function's object
        :param factory_ctx: (dict) names available for Function instantiation
        :return: an object that implements a check function
        """
        return propartial(JoblocksWrapper, cmdb_url=self.cmdb_url, project=factory_ctx['entity'].get('name'))


class JoblocksWrapper(object):

    LOCKING_NODE_ROLE_ID = 118
    ALLOCATED_STATUS_ID = 6000
    DEFAULT_EXPECTED_DURATION = 60000  # [ms]

    def __init__(self, cmdb_url, project=None, environment='LIVE'):
        if cmdb_client:
            self._cmdb = cmdb_client(cmdb_url)
        self.pattern = 'job:lock:{}:{}:*'.format(project or '*', environment)

    @memory_cache.cache_on_arguments(namespace='zmon-worker')
    def _get_hosts(self, host_role_id, lifecycle_status_id):
        return self._cmdb.get_hosts(host_role_id=host_role_id, lifecycle_status_id=lifecycle_status_id)

    @staticmethod
    def _get_expected_duration(redis_value, check_param):
        '''
        >>> JoblocksWrapper._get_expected_duration({}, None)
        60000.0
        >>> JoblocksWrapper._get_expected_duration({}, 10000)
        10000.0
        >>> JoblocksWrapper._get_expected_duration({'expectedMaximumDuration': 120000}, None)
        120000.0
        >>> JoblocksWrapper._get_expected_duration({'expectedMaximumDuration': 60000}, 90000)
        90000.0
        '''

        return float((check_param if check_param else redis_value.get('expectedMaximumDuration',
                     JoblocksWrapper.DEFAULT_EXPECTED_DURATION)))

    def results(self, expected_duration=None):
        hosts = self._get_hosts(JoblocksWrapper.LOCKING_NODE_ROLE_ID, JoblocksWrapper.ALLOCATED_STATUS_ID)
        host_connections = dict((host.hostname, redis.StrictRedis(host=host.hostname)) for host in hosts)
        host_keys = dict((host, con.keys(self.pattern)) for (host, con) in host_connections.iteritems())
        str_results = []

        for host, keys in host_keys.iteritems():
            p = host_connections[host].pipeline()
            for key in keys:
                p.get(key)
            str_results.extend(p.execute())

        results = []
        for r in str_results:
            try:
                results.append(json.loads(r))
            except ValueError:
                pass

        # In case flowId is None, we want to return empty string instead.
        return dict((r['lockingComponent'], {
            'host': r['host'],
            'instance': r['instanceCode'],
            'created': r['created'],
            'expected_duration': JoblocksWrapper._get_expected_duration(r, expected_duration),
            'flow_id': r.get('flowId') or '',
            'expired': time.time() - time.mktime(time.strptime(r['created'], '%Y-%m-%dT%H:%M:%S'))
                > JoblocksWrapper._get_expected_duration(r, expected_duration) / 1000,
        }) for r in results)


if __name__ == '__main__':
    import sys

    if len(sys.argv) != 2:
        print "usage: {} <cmdb_url>".format(sys.argv[0])
        sys.exit(1)

    res = JoblocksWrapper(cmdb_url=sys.argv[1], environment='LIVE').results(expected_duration=60000)
    print res



# -*- coding: utf-8 -*-
"""
Logic for controlling worker processes
"""

import os
import signal
import time
import copy
import logging

from multiprocessing import Process
from threading import Thread

logger = logging.getLogger(__name__)


class ProcessController(object):
    """
    Class to handle a bunch of child processes
    what can it do:
    0. define a common target function for every process?
    1. spawn N processes that execute the target function, store references to objects and its pid
    2. spawn more process after some are running
    3. terminate some process *(by pid?)
    4. spawn a thread loop for checking the health of child processes *(and take some action if some process dies)?
    5. dynamically change the policy on how to react to process dies *(use queue for incoming requests?)
    """

    action_policies = ('report', 'dynamic_num', 'dynamic_throughput')

    proc_stat_element = {'begin_time': -1, 'end_time': -1, 'alive': False, 'rebel': False, 'pid': 0, 'exitcode': 0,
                         'mem': -1, 'abnormal_termination': False}

    def __init__(self, default_target=None, default_args=None, default_kwargs=None, always_add_kwargs=None,
                 action_policy='report', min_processes=2, max_processes=1000, start_action_loop=True):

        self.default_target = default_target
        self.default_args = default_args if isinstance(default_args, (tuple, list)) else ()
        self.default_kwargs = default_kwargs if isinstance(default_kwargs, dict) else {}
        self.always_add_kwargs = always_add_kwargs if isinstance(always_add_kwargs, dict) else {}

        self.proc_dict = {}    # { proc_name : proc}

        self.proc_rebel = {}   # { proc_name : proc} -> for processes that refuse to die
        self.proc_stats = {}   # { proc_name : {'begin_time':T0, 'end_time': T1, 'alive' : bool1, ... } }

        self.proc_args = {}    # { proc_name : {'target':None, 'args':[...], 'kwargs'={...} }}

        self.pid_to_pname = {}   # {pid: proc_name}

        self.pids_for_termination = []  # [ pid1, pid2, ....]

        self.limbo_proc_dict = {}       # {proc_name : proc}

        self.max_killed_stats = 5000         # max number of dead proc stats to keep around in memory
        self.min_processes = min_processes
        self.max_processes = max_processes
        self.count_stop_condition = 0        # counter of consecutive stop conditions found
        self.consecutive_stop_condition = 5        # counter of consecutive stop conditions found
        self.gracetime_stop_condition = 60   # number of seconds to wait before a final stop condition check

        self._thread_action_loop = None
        self.stop_action = True
        self.action_loop_interval = 2      # seconds between each actions pass
        self.set_action_policy(action_policy)
        self.set_dynamic_num_processes(5)  # number of process to maintain alive when action_policy == 'dynamic_num'

        self._tstamp_clean_old_proc_stats = -1    # timestamp of the last execution of _clean_old_proc_stats()
        self._tdelta_clean_old_proc_stats = 300   # frequency of __clean_old_proc_stats()

        if start_action_loop:
            self.start_action_loop()

    def spawn_process(self, target=None, args=None, kwargs=None):

        args = args if isinstance(args, (tuple, list)) else ()
        kwargs = kwargs if isinstance(kwargs, dict) else {}

        if self.max_processes == len(self.proc_dict):
            raise Exception("maximum number of processes reached!!!")

        target = target or self.default_target
        args = args or self.default_args
        kwargs = dict(kwargs if kwargs else self.default_kwargs)
        kwargs.update(self.always_add_kwargs)

        try:
            proc = Process(target=target, args=args, kwargs=kwargs)
            proc.start()
            pname = proc.name

            # creating entry in running process table
            self.proc_dict[pname] = proc

            # mapping pid -> pname
            self.pid_to_pname[proc.pid] = pname

            # store process arguments to relaunch it if it dies
            self.proc_args[pname] = dict(target=target, args=args, kwargs=kwargs)

            # creating entry in stats table
            self.proc_stats[pname] = dict(self.proc_stat_element)
            self.proc_stats[pname]['pid'] = proc.pid
            self.proc_stats[pname]['alive'] = proc.is_alive()
            self.proc_stats[pname]['begin_time'] = time.time()    # use self._format_time() to get datetime format

        except Exception:
            logger.exception("Spawn of process failed. Caught exception with details: ")
            raise

        return pname

    def spawn_many(self, N, target=None, args=None, kwargs=None):

        logger.info('>>>>>>> spawn_many: %d, %s, %s', N, args, kwargs)

        args = args if isinstance(args, (tuple, list)) else ()
        kwargs = kwargs if isinstance(kwargs, dict) else {}

        n_success = 0
        for i in range(N):
            try:
                self.spawn_process(target=target, args=args, kwargs=kwargs)
            except Exception:
                logger.exception('Failed to start process. Reason: ')
            else:
                n_success += 1
        return n_success == N

    def terminate_process(self, proc_name, kill_wait=0.5):

        proc = self.proc_dict.get(proc_name)
        if not proc:
            logger.warn('process: %s not found!!!!!!!!!!!!!!!!!', proc_name)
            return False

        if proc.is_alive():
            logger.warn('terminating process: %s', proc_name)
            proc.terminate()

            time.sleep(kill_wait)

            if proc.is_alive():
                logger.warn('Sending SIGKILL to process with pid=%s', proc.pid)
                os.kill(proc.pid, signal.SIGKILL)

            abnormal_termination = False
        else:
            logger.warn('process: %s is not alive!!!!!!!!!!!!!!!!!', proc_name)
            abnormal_termination = True

        # move proc to limbo and record end time in stats
        self.proc_dict.pop(proc_name, None)
        self.limbo_proc_dict[proc_name] = proc
        self._close_proc_stats(proc, abnormal_termination)
        return True

    def terminate_all_processes(self):

        self.stop_action_loop()   # very important: stop action loop before starting to terminate child processes

        all_pnames = list(self.proc_dict.keys())

        for proc_name in all_pnames:
            self.terminate_process(proc_name,  kill_wait=0.1)

        logger.info("proc_stats after terminate_all_processes() : %s", self.list_stats())
        return True

    def _close_proc_stats(self, proc, abnormal_termination=False):
        # Update proc_stats  {'proc_name' : {'begin_time':T0, 'end_time': T1, 'alive' : bool1,... } }
        pn = proc.name
        if proc.is_alive():
            self.proc_stats[pn]['alive'] = True
            self.proc_stats[pn]['rebel'] = True
        else:
            self.proc_stats[pn]['abnormal_termination'] = abnormal_termination
            self.proc_stats[pn]['end_time'] = time.time()
            self.proc_stats[pn]['alive'] = False
            self.proc_stats[pn]['exitcode'] = proc.exitcode

    def get_info(self, proc_name):
        """
        Get all the info I can of this process, for example:
        1. How long has it been running? *(Do I need an extra pid table for statistics?)
        2. How much memory does it use?
        """
        raise NotImplementedError('Method get_info not implemented yet')

    def list_running(self):
        return [(proc_name, proc.pid) for proc_name, proc in self.proc_dict.items()]

    def list_stats(self):
        proc_stats = copy.deepcopy(self.proc_stats)
        for proc_name, stats in proc_stats.items():
            stats['begin_time'] = self._format_time(stats['begin_time'])
            stats['end_time'] = self._format_time(stats['end_time'])
        return proc_stats

    def _format_time(self, seconds):
        return time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(seconds)) if seconds else '--'

    def get_action_policy(self):
        return self.action_policy

    def set_action_policy(self, action_policy):
        if action_policy not in self.action_policies:
            raise Exception('Invalid action policy, possible values are: %s' % ', '.join(self.action_policies))
        self.action_policy = action_policy

    def available_action_policies(self):
        return self.action_policies

    def is_action_loop_running(self):
        return not self.stop_action

    def get_dynamic_num_processes(self):
        return self.dynamic_num_processes

    def set_dynamic_num_processes(self, dynamic_num_processes):
        try:
            assert type(dynamic_num_processes) is int and \
                   self.min_processes <= dynamic_num_processes <= self.max_processes
        except AssertionError:
            raise Exception('dynamic_num_processes passed is not in correct range')
        self.dynamic_num_processes = dynamic_num_processes

    def _clean_old_proc_stats(self):
        """ Remove old stats from dead processes to avoid high memory usage """
        if time.time() - self._tstamp_clean_old_proc_stats > self._tdelta_clean_old_proc_stats:
            self._tstamp_clean_old_proc_stats = time.time()
            et_pn = sorted([(stats['end_time'], pn) for pn, stats in self.proc_stats.copy().items() if stats['end_time'] > 0])
            del_et_pn = et_pn[:len(et_pn)-self.max_killed_stats] if len(et_pn) > self.max_killed_stats else []
            for end_time, pn in del_et_pn:
                stats = self.proc_stats.pop(pn, None)
                logger.warn('Deleting stats of killed process %s to preserve memory: %s', pn, stats)

    def _clean_limbo_procs(self):

        limbo_dict = dict(self.limbo_proc_dict)

        for pname, proc in limbo_dict.items():
            if proc.is_alive():
                logger.error('Fatal: process in limbo in undead state!!!!!')
            else:
                self.pid_to_pname.pop(proc.pid, None)
                self.proc_args.pop(pname, None)
                self.limbo_proc_dict.pop(pname, None)

    def mark_for_termination(self, pid):
        """Given pid will be stored in local variable that marks them for termination in the next action pass"""
        self.pids_for_termination.append(pid)

    def _respawn(self, proc_name):
        # terminate process and spawn another process with same arguments
        pargs = self.proc_args.get(proc_name, {})
        proc = self.proc_dict.get(proc_name)
        pid = proc.pid if proc else '???'
        was_alive = proc.is_alive() if proc else '???'
        self.terminate_process(proc_name, kill_wait=1.0)
        proc_name2 = self.spawn_process(**pargs)
        proc2 = self.proc_dict.get(proc_name2)
        pid2 = proc2.pid if proc2 else '???'
        logger.warn('Respawned process: proc_name=%s, pid=%s, was_alive=%s --> proc_name=%s, pid=%s, args=%s', proc_name, pid, was_alive, proc_name2, pid2, pargs)

    def _action_loop(self):
        """
        A threaded loop that runs every interval seconds to perform autonomous actions
        """
        while not self.stop_action:
            try:
                # action 1: respond to kill requests: terminate marked pid and spawn them again
                term_pids = self.pids_for_termination[:]
                del self.pids_for_termination[:len(term_pids)]

                for pid in term_pids:
                    pname = self.pid_to_pname.get(pid)
                    if pname is not None:
                        if not self.stop_action:
                            self._respawn(pname)
                    else:
                        logger.warn('action_loop found non valid pid: %s', pid)

                # action 2: inspect all processes and react to those that died unexpectedly
                for pname, proc in self.proc_dict.items():
                    if not proc.is_alive() and not self.stop_action:
                        logger.warn('Detected abnormal termination of process pid=%s... attempting restart', proc.pid)
                        self._respawn(pname)

                self._clean_limbo_procs()
                self._clean_old_proc_stats()
            except Exception:
                logger.exception('Error in ProcessController action_loop: ')

            if not self.stop_action:
                time.sleep(self.action_loop_interval)

    def start_action_loop(self, interval=1):
        self.stop_action = False
        self.action_loop_interval = interval

        self._thread_action_loop = Thread(target=self._action_loop)
        self._thread_action_loop.daemon = True
        self._thread_action_loop.start()

    def stop_action_loop(self):
        self.stop_action = True
